IF EXISTS 
(
  SELECT name FROM sysobjects
  WHERE name = 'Object_TraceModifications' AND type = 'TR'
)
DROP TRIGGER Object_TraceModifications

GO

CREATE TRIGGER [dbo].[Object_TraceModifications]
   ON  [dbo].[Object] 
   AFTER INSERT,DELETE,UPDATE
AS 
BEGIN
  SET NOCOUNT ON

  DECLARE @SetupPresent INTEGER;
  DECLARE @SetupTriggerTestStatus INTEGER;
  DECLARE @SetupSQLCheckObjectLockType INTEGER;
  DECLARE @SetupTraceModifications INTEGER;
  DECLARE @SkipSQLTrigger INTEGER;
  DECLARE @DatabaseLocked INTEGER;
  DECLARE @Username NVARCHAR(100);
  DECLARE @LockedBy NVARCHAR(100);
  DECLARE @IsLocked INTEGER;
  DECLARE @IsModification INTEGER;
  DECLARE @CheckIsLocked INTEGER;
  DECLARE @CalledFromRepository INTEGER;
  DECLARE @ObjectType INTEGER;
  DECLARE @ObjectId INTEGER;
  DECLARE @ObjectName NVARCHAR(100);
  DECLARE @ObjectDate DATETIME
  DECLARE @ObjectTime DATETIME
  DECLARE @ObjectTypeText NVARCHAR(100);
  DECLARE @ObjectIdText NVARCHAR(100);
  DECLARE @Message NVARCHAR(100);
  DECLARE @NoOfInserts INTEGER;
  DECLARE @NoOfDeletes INTEGER;
  DECLARE @NoOfModifies INTEGER;
  DECLARE @Action INTEGER;
  DECLARE @TokenNo INTEGER;
  DECLARE @LockObjectAtSaving INTEGER;
  DECLARE @OldVersionList NVARCHAR(80);
  DECLARE @NewVersionList NVARCHAR(100);

  SELECT TOP 1
    @SetupPresent = 1,
    @SetupTriggerTestStatus = [Trigger Test Status],
    @SetupSQLCheckObjectLockType = [SQL Check Object Lock Type],
    @SetupTraceModifications = [Trace Modifications],
    @SkipSQLTrigger = [Skip SQL Trigger],
    @LockObjectAtSaving = [Lock Object at Saving],
    @DatabaseLocked = [Database Locked]
  FROM [OM - Setup] 
  WHERE [Primary Key] = '';

  IF @SetupPresent = 1
  BEGIN

    IF @SetupTriggerTestStatus = 1
    BEGIN
      UPDATE [OM - Setup] 
      SET [Trigger Test Status] = 4;
    END ELSE BEGIN

      IF @SkipSQLTrigger = 0 
      BEGIN

        IF @DatabaseLocked = 1 
        BEGIN
          RAISERROR('Database is locked.', 16, 1);
          ROLLBACK TRANSACTION;
        END;

        SELECT
          @ObjectType = d.[Type],
          @ObjectId = d.[ID],
          @ObjectName = d.[Name],
          @ObjectDate = d.[Date],
          @ObjectTime = d.[Time]
        FROM Deleted d;

        SELECT
          @ObjectType = i.[Type],
          @ObjectId = i.[ID],
          @ObjectName = i.[Name],
          @ObjectDate = i.[Date],
          @ObjectTime = i.[Time]
        FROM Inserted i;
        
        SET @CalledFromRepository = 0;
        SELECT @CalledFromRepository = 1
        FROM [OM - Repository Log] rl
        WHERE rl.[Object Type] = @ObjectType
        AND rl.[Object No_] = @ObjectId
        AND rl.Status = 1;

        IF @ObjectType > 0 AND (@ObjectId < 1000000000 OR @ObjectId >= 2000000000) AND @CalledFromRepository = 0 
        BEGIN
          SET @LockedBy = '';
          SET @TokenNo = 0;
          SET @IsLocked = 0;

          SELECT
            @LockedBy = ol.[Locked By],
            @LockedBy = ol.[Locked By],
            @TokenNo = ol.[Token No_],
            @IsLocked = 1
          FROM [OM - Object Lock] ol
          WHERE ol.[Object Type] = @ObjectType
          AND ol.[Object No_] = @ObjectId;

          SET @Username = UPPER(SYSTEM_USER);
          IF (CHARINDEX('\', @Username) > 0)
            SET @Username = SUBSTRING(@Username, CHARINDEX('\', @Username) + 1, 100);

          SELECT @NoOfInserts = COUNT([Type]) FROM Inserted;
          SELECT @NoOfDeletes = COUNT([Type]) FROM Deleted;
          SELECT @NoOfModifies = COUNT(i.[Type])
          FROM Inserted i 
            INNER JOIN Deleted d 
              ON i.[Type] = d.[Type] AND i.[ID] = d.[ID];

          SET @IsModification = 0;
          SET @CheckIsLocked = 0;

          -- INSERT
          IF @NoOfModifies = 0 AND @NoOfInserts > 0 
          BEGIN
            SET @Action = 1;
            SET @IsModification = 1;
          END;

          -- MODIFY
          IF @NoOfModifies > 0
            SET @Action = 2;

          -- DELETE
          IF @NoOfDeletes > 0 AND @NoOfInserts = 0 
          BEGIN
            SET @Action = 3;
            SET @IsModification = 1;
            SET @CheckIsLocked = 1;
          END;

          -- RENAME
          IF @NoOfModifies = 0 AND @NoOfDeletes > 0 AND @NoOfInserts > 0 
          BEGIN
            SET @Action = 4
            SET @IsModification = 1;
            SELECT @Action = 0, @IsModification = 0
            FROM Inserted i, [OM - Update Object] uo
            WHERE i.[Type] = uo.[Object Type]
            AND i.[ID] = uo.[New Object No_];
          END;

          -- MODIFY
          IF @Action = 2
            SELECT 
              @IsModification = 1,
              @CheckIsLocked = 1
            FROM Inserted i
              INNER JOIN Deleted d
                ON i.[Type] = d.[Type]
                AND i.[ID] = d.[ID]
            WHERE 
            (
              i.[Date] <> d.[Date]
              OR CONVERT(VARCHAR(20), i.[Time], 108) <> CONVERT(VARCHAR(20), d.[Time], 108)
              OR i.[Name] <> d.[Name]
              OR REPLACE(REPLACE(REPLACE(i.[Version List], ',', ''), '#', ''), 'LOCKED', '') <>
                 REPLACE(REPLACE(REPLACE(d.[Version List], ',', ''), '#', ''), 'LOCKED', '')
            )
            AND i.[Version List] <> '! CHECK OBJECT VALID !'
            AND d.[Version List] <> '! CHECK OBJECT VALID !';
          
          IF @SetupSQLCheckObjectLockType <> 0 AND @CheckIsLocked = 1 AND 
             @LockedBy <> @Username AND @LockedBy <> ''
          BEGIN
            SELECT 
              @ObjectTypeText = 
                CASE @ObjectType
                  WHEN 1 THEN 'Table'
                  WHEN 2 THEN 'Form'
                  WHEN 3 THEN 'Report'
                  WHEN 4 THEN 'Dataport'
                  WHEN 5 THEN 'Codeunit'
                  WHEN 6 THEN 'XMLport'
                  WHEN 7 THEN 'MenuSuite'
                  WHEN 8 THEN 'Page'
                  ELSE ''
                END,
              @ObjectIdText = @ObjectId;

            SET @Message = 
              CHAR(13) + CHAR(10) + CHAR(13) + CHAR(10) + 
              'OBJECT MANAGER ERROR:' + CHAR(13) + CHAR(10) +
              'Object %s %s - %s is locked by %s' + CHAR(13) + CHAR(10);
            RAISERROR(@Message, 16, 1, @ObjectTypeText, @ObjectIdText, @ObjectName, @LockedBy);
            ROLLBACK TRANSACTION;
          END;

          IF @LockObjectAtSaving = 1 AND @IsModification = 1 AND @LockedBy = '' AND @Action <> 4
          BEGIN
            
            IF EXISTS(
              SELECT 1 
              FROM [OM - Repository Setup]
              WHERE [Use Repository] = 1)
            BEGIN
              SET @Message = 
                CHAR(13) + CHAR(10) + CHAR(13) + CHAR(10) + 
                'OBJECT MANAGER ERROR:' + CHAR(13) + CHAR(10) +
                'The option ''Lock Object at Saving'' cannot be ' + 
                'used in combination with repository' + CHAR(13) + CHAR(10);
              RAISERROR(@Message, 16, 1);
              ROLLBACK TRANSACTION;
            END;

            INSERT INTO [OM - Object Lock]
              (
                [Object Type], [Object No_], [Locked By], 
                [Lock Date], 
                [Lock Time], 
                [Deleted], [Token No_], [Branch No_]
              )
            SELECT 
              @ObjectType, @ObjectId, @Username, 
              CAST(CONVERT(VARCHAR(20), GETDATE(), 112) + ' 00:00:00' AS DATETIME), 
              CAST('17540101 ' + CONVERT(VARCHAR(20), GETDATE(), 108) AS DATETIME),
              0, 0, '';
            SELECT @OldVersionList = [Version List] 
            FROM Inserted;

            IF CHARINDEX('#', @OldVersionList) = 1
              SET @NewVersionList = '#LOCKED,' + SUBSTRING(@OldVersionList, 2, 100)
            ELSE
              SET @NewVersionList = 'LOCKED,' + @OldVersionList;

            IF LEN(@NewVersionList) <= 80 
              UPDATE [Object] 
              SET [Version List] = @NewVersionList
              WHERE [Object].[Type] = @ObjectType
              AND [Object].[ID] = @ObjectId
              AND CHARINDEX('LOCKED', [Version List]) = 0;

            SET @IsLocked = 1;
            SET @LockedBy = @Username;

          END;

          IF @SetupSQLCheckObjectLockType = 2 AND @CheckIsLocked = 1 AND @IsLocked = 0
          BEGIN
            SELECT 
              @ObjectTypeText = 
                CASE @ObjectType
                  WHEN 1 THEN 'Table'
                  WHEN 2 THEN 'Form'
                  WHEN 3 THEN 'Report'
                  WHEN 4 THEN 'Dataport'
                  WHEN 5 THEN 'Codeunit'
                  WHEN 6 THEN 'XMLport'
                  WHEN 7 THEN 'MenuSuite'
                  WHEN 8 THEN 'Page'
                  ELSE ''
                END,
              @ObjectIdText = @ObjectId;

            SET @Message = 
              CHAR(13) + CHAR(10) + CHAR(13) + CHAR(10) + 
              'OBJECT MANAGER ERROR:' + CHAR(13) + CHAR(10) +
              'Object %s %s - %s is not locked' + CHAR(13) + CHAR(10);
            RAISERROR(@Message, 16, 1, @ObjectTypeText, @ObjectIdText, @ObjectName);
            ROLLBACK TRANSACTION;
          END; 

          IF @SetupTraceModifications = 1 AND @IsModification = 1
            INSERT INTO [OM - Modification] 
              (
                [Object Type], [Object No_], [Object Name], 
                [Object Date], [Object Time], 
                [Status], [Inserted By], 
                [Insert Date], 
                [Insert Time],
                [Assigned to Project No_], [Assigned By],
                [Assign Date], [Assign Time], [Auto Assigned], 
                [Transport No_], [Object Date Time],
                [Locked By], [Token No_], [Traced By SQL], [SQL Trigger], [SQL Status], 
                [System User], [Host Name]
              )
            SELECT 
              @ObjectType, @ObjectId, @ObjectName, 
              @ObjectDate, CAST('17540101 ' + CONVERT(VARCHAR(20), @ObjectTime, 108) AS DATETIME),
              0, @Username, 
              CAST(CONVERT(VARCHAR(20), GETDATE(), 112) + ' 00:00:00' AS DATETIME), 
              CAST('17540101 ' + CONVERT(VARCHAR(20), GETDATE(), 108) AS DATETIME),
              '', '', 
              CAST('17530101 00:00:00' AS DATETIME), CAST('17530101 00:00:00' AS DATETIME), 0,
              '', CAST(CONVERT(VARCHAR(20), @ObjectDate, 112) + ' ' + 
              CONVERT(VARCHAR(20), @ObjectTime, 108) AS DATETIME),
              @LockedBy, @TokenNo, 1, @Action, 1, 
              SUBSTRING(SYSTEM_USER, 1, 50), SUBSTRING(HOST_NAME(), 1, 50);

        END;  

      END;
    END;

  END;

END;
